@@ -1,24 +1,35 @@
 import logging
 import warnings
-from typing import Optional, Sequence, Union
+from copy import deepcopy
+from typing import List, Optional, Sequence, Union
 
 import numpy as np
 import pandas as pd
 import torch
 from anndata import AnnData
-from pandas.api.types import CategoricalDtype
 
-from scvi import _CONSTANTS
+from scvi import REGISTRY_KEYS
 from scvi._compat import Literal
-from scvi.data._anndata import _make_obs_column_categorical
+from scvi.data import AnnDataManager
+from scvi.data._constants import _SETUP_ARGS_KEY
+from scvi.data.fields import (
+    CategoricalJointObsField,
+    CategoricalObsField,
+    LabelsWithUnlabeledObsField,
+    LayerField,
+    NumericalJointObsField,
+    NumericalObsField,
+)
 from scvi.dataloaders import (
     AnnDataLoader,
     SemiSupervisedDataLoader,
     SemiSupervisedDataSplitter,
 )
+from scvi.model._utils import _init_library_size
 from scvi.module import SCANVAE
 from scvi.train import SemiSupervisedTrainingPlan, TrainRunner
 from scvi.train._callbacks import SubSampleLabels
+from scvi.utils import setup_anndata_dsp
 
 from ._scvi import SCVI
 from .base import ArchesMixin, BaseModelClass, RNASeqMixin, VAEMixin
@@ -35,9 +46,7 @@ class SCANVI(RNASeqMixin, VAEMixin, ArchesMixin, BaseModelClass):
     Parameters
     ----------
     adata
-        AnnData object that has been registered via :func:`~scvi.data.setup_anndata`.
-    unlabeled_category
-        Value used for unlabeled cells in `labels_key` used to setup AnnData with scvi.
+        AnnData object that has been registered via :meth:`~scvi.model.SCANVI.setup_anndata`.
     n_hidden
         Number of nodes per hidden layer.
     n_latent
@@ -65,7 +74,7 @@ class SCANVI(RNASeqMixin, VAEMixin, ArchesMixin, BaseModelClass):
     Examples
     --------
     >>> adata = anndata.read_h5ad(path_to_anndata)
-    >>> scvi.data.setup_anndata(adata, batch_key="batch", labels_key="labels")
+    >>> scvi.model.SCANVI.setup_anndata(adata, batch_key="batch", labels_key="labels")
     >>> vae = scvi.model.SCANVI(adata, "Unknown")
     >>> vae.train()
     >>> adata.obsm["X_scVI"] = vae.get_latent_representation()
@@ -75,15 +84,14 @@ class SCANVI(RNASeqMixin, VAEMixin, ArchesMixin, BaseModelClass):
     -----
     See further usage examples in the following tutorials:
 
-    1. :doc:`/user_guide/notebooks/harmonization`
-    2. :doc:`/user_guide/notebooks/scarches_scvi_tools`
-    3. :doc:`/user_guide/notebooks/seed_labeling`
+    1. :doc:`/tutorials/notebooks/harmonization`
+    2. :doc:`/tutorials/notebooks/scarches_scvi_tools`
+    3. :doc:`/tutorials/notebooks/seed_labeling`
     """
 
     def __init__(
         self,
         adata: AnnData,
-        unlabeled_category: Union[str, int, float],
         n_hidden: int = 128,
         n_latent: int = 10,
         n_layers: int = 1,
@@ -95,8 +103,7 @@ class SCANVI(RNASeqMixin, VAEMixin, ArchesMixin, BaseModelClass):
         super(SCANVI, self).__init__(adata)
         scanvae_model_kwargs = dict(model_kwargs)
 
-        self.unlabeled_category_ = unlabeled_category
-        has_unlabeled = self._set_indices_and_labels()
+        self._set_indices_and_labels()
 
         if len(self._labeled_indices) != 0:
             self._dl_cls = SemiSupervisedDataLoader
@@ -104,21 +111,30 @@ class SCANVI(RNASeqMixin, VAEMixin, ArchesMixin, BaseModelClass):
             self._dl_cls = AnnDataLoader
 
         # ignores unlabeled catgegory
-        n_labels = (
-            self.summary_stats["n_labels"] - 1
-            if has_unlabeled
-            else self.summary_stats["n_labels"]
-        )
+        n_labels = self.summary_stats.n_labels - 1
         n_cats_per_cov = (
-            self.scvi_setup_dict_["extra_categoricals"]["n_cats_per_key"]
-            if "extra_categoricals" in self.scvi_setup_dict_
+            self.adata_manager.get_state_registry(
+                REGISTRY_KEYS.CAT_COVS_KEY
+            ).n_cats_per_key
+            if REGISTRY_KEYS.CAT_COVS_KEY in self.adata_manager.data_registry
             else None
         )
+
+        n_batch = self.summary_stats.n_batch
+        use_size_factor_key = (
+            REGISTRY_KEYS.SIZE_FACTOR_KEY in self.adata_manager.data_registry
+        )
+        library_log_means, library_log_vars = None, None
+        if not use_size_factor_key:
+            library_log_means, library_log_vars = _init_library_size(
+                self.adata_manager, n_batch
+            )
+
         self.module = SCANVAE(
-            n_input=self.summary_stats["n_vars"],
-            n_batch=self.summary_stats["n_batch"],
+            n_input=self.summary_stats.n_vars,
+            n_batch=n_batch,
             n_labels=n_labels,
-            n_continuous_cov=self.summary_stats["n_continuous_covs"],
+            n_continuous_cov=self.summary_stats.get("n_extra_continuous_covs", 0),
             n_cats_per_cov=n_cats_per_cov,
             n_hidden=n_hidden,
             n_latent=n_latent,
@@ -126,6 +142,9 @@ class SCANVI(RNASeqMixin, VAEMixin, ArchesMixin, BaseModelClass):
             dropout_rate=dropout_rate,
             dispersion=dispersion,
             gene_likelihood=gene_likelihood,
+            use_size_factor_key=use_size_factor_key,
+            library_log_means=library_log_means,
+            library_log_vars=library_log_vars,
             **scanvae_model_kwargs,
         )
 
@@ -136,7 +155,7 @@ class SCANVI(RNASeqMixin, VAEMixin, ArchesMixin, BaseModelClass):
             "ScanVI Model with the following params: \nunlabeled_category: {}, n_hidden: {}, n_latent: {}"
             ", n_layers: {}, dropout_rate: {}, dispersion: {}, gene_likelihood: {}"
         ).format(
-            unlabeled_category,
+            self.unlabeled_category_,
             n_hidden,
             n_latent,
             n_layers,
@@ -152,25 +171,31 @@ class SCANVI(RNASeqMixin, VAEMixin, ArchesMixin, BaseModelClass):
         cls,
         scvi_model: SCVI,
         unlabeled_category: str,
+        labels_key: Optional[str] = None,
         adata: Optional[AnnData] = None,
         **scanvi_kwargs,
     ):
         """
-        Initialize scanVI model with weights from pretrained scVI model.
+        Initialize scanVI model with weights from pretrained :class:`~scvi.model.SCVI` model.
 
         Parameters
         ----------
         scvi_model
             Pretrained scvi model
+        labels_key
+            key in `adata.obs` for label information. Label categories can not be different if
+            labels_key was used to setup the SCVI model. If None, uses the `labels_key` used to
+            setup the SCVI model. If that was None, and error is raised.
         unlabeled_category
             Value used for unlabeled cells in `labels_key` used to setup AnnData with scvi.
         adata
-            AnnData object that has been registered via :func:`~scvi.data.setup_anndata`.
+            AnnData object that has been registered via :meth:`~scvi.model.SCANVI.setup_anndata`.
         scanvi_kwargs
-            kwargs for scanVI model
+            kwargs for scANVI model
         """
-        if scvi_model.is_trained_ is False:
-            warnings.warn("Passed in scvi model hasn't been trained yet.")
+        scvi_model._check_if_trained(
+            message="Passed in scvi model hasn't been trained yet."
+        )
 
         scanvi_kwargs = dict(scanvi_kwargs)
         init_params = scvi_model.init_params_
@@ -187,10 +212,24 @@ class SCANVI(RNASeqMixin, VAEMixin, ArchesMixin, BaseModelClass):
 
         if adata is None:
             adata = scvi_model.adata
+        else:
+            # validate new anndata against old model
+            scvi_model._validate_anndata(adata)
 
-        scanvi_model = cls(
-            adata, unlabeled_category, **non_kwargs, **kwargs, **scanvi_kwargs
+        scvi_setup_args = deepcopy(scvi_model.adata_manager.registry[_SETUP_ARGS_KEY])
+        scvi_labels_key = scvi_setup_args["labels_key"]
+        if labels_key is None and scvi_labels_key is None:
+            raise ValueError(
+                "A `labels_key` is necessary as the SCVI model was initialized without one."
+            )
+        if scvi_labels_key is None:
+            scvi_setup_args.update(dict(labels_key=labels_key))
+        cls.setup_anndata(
+            adata,
+            unlabeled_category=unlabeled_category,
+            **scvi_setup_args,
         )
+        scanvi_model = cls(adata, **non_kwargs, **kwargs, **scanvi_kwargs)
         scvi_state_dict = scvi_model.module.state_dict()
         scanvi_model.module.load_state_dict(scvi_state_dict, strict=False)
         scanvi_model.was_pretrained = True
@@ -199,48 +238,23 @@ class SCANVI(RNASeqMixin, VAEMixin, ArchesMixin, BaseModelClass):
 
     def _set_indices_and_labels(self):
         """
-        Set indices and make unlabeled cat as the last cat.
-
-        Returns
-        -------
-        True is categories reordered else False
+        Set indices for labeled and unlabeled cells.
         """
-        # get indices for labeled and unlabeled cells
-        key = self.scvi_setup_dict_["data_registry"][_CONSTANTS.LABELS_KEY]["attr_key"]
-        mapping = self.scvi_setup_dict_["categorical_mappings"][key]["mapping"]
-        original_key = self.scvi_setup_dict_["categorical_mappings"][key][
-            "original_key"
-        ]
-        labels = np.asarray(self.adata.obs[original_key]).ravel()
+        labels_state_registry = self.adata_manager.get_state_registry(
+            REGISTRY_KEYS.LABELS_KEY
+        )
+        self.unlabeled_category_ = labels_state_registry.unlabeled_category
 
-        if self.unlabeled_category_ in labels:
-            unlabeled_idx = np.where(mapping == self.unlabeled_category_)
-            unlabeled_idx = unlabeled_idx[0][0]
-            # move unlabeled category to be the last position
-            mapping[unlabeled_idx], mapping[-1] = mapping[-1], mapping[unlabeled_idx]
-            cat_dtype = CategoricalDtype(categories=mapping, ordered=True)
-            # rerun setup for the batch column
-            _make_obs_column_categorical(
-                self.adata,
-                original_key,
-                "_scvi_labels",
-                categorical_dtype=cat_dtype,
-            )
-            remapped = True
-        else:
-            remapped = False
+        labels = self.get_from_registry(self.adata, REGISTRY_KEYS.LABELS_KEY)
+        self._label_mapping = labels_state_registry.categorical_mapping
 
-        self.scvi_setup_dict_ = self.adata.uns["_scvi"]
-        self._label_mapping = mapping
         # set unlabeled and labeled indices
         self._unlabeled_indices = np.argwhere(
             labels == self.unlabeled_category_
         ).ravel()
         self._labeled_indices = np.argwhere(labels != self.unlabeled_category_).ravel()
         self._code_to_label = {i: l for i, l in enumerate(self._label_mapping)}
-        self.original_label_key = original_key
-
-        return remapped
+        self.original_label_key = labels_state_registry.original_key
 
     def predict(
         self,
@@ -255,7 +269,7 @@ class SCANVI(RNASeqMixin, VAEMixin, ArchesMixin, BaseModelClass):
         Parameters
         ----------
         adata
-            AnnData object that has been registered via :func:`~scvi.data.setup_anndata`.
+            AnnData object that has been registered via :meth:`~scvi.model.SCANVI.setup_anndata`.
         indices
             Return probabilities for each class label.
         soft
@@ -275,8 +289,8 @@ class SCANVI(RNASeqMixin, VAEMixin, ArchesMixin, BaseModelClass):
         )
         y_pred = []
         for _, tensors in enumerate(scdl):
-            x = tensors[_CONSTANTS.X_KEY]
-            batch = tensors[_CONSTANTS.BATCH_KEY]
+            x = tensors[REGISTRY_KEYS.X_KEY]
+            batch = tensors[REGISTRY_KEYS.BATCH_KEY]
             pred = self.module.classify(x, batch)
             if not soft:
                 pred = pred.argmax(dim=1)
@@ -358,7 +372,7 @@ class SCANVI(RNASeqMixin, VAEMixin, ArchesMixin, BaseModelClass):
         )
 
         data_splitter = SemiSupervisedDataSplitter(
-            adata=self.adata,
+            adata_manager=self.adata_manager,
             unlabeled_category=self.unlabeled_category_,
             train_size=train_size,
             validation_size=validation_size,
@@ -382,3 +396,52 @@ class SCANVI(RNASeqMixin, VAEMixin, ArchesMixin, BaseModelClass):
             **trainer_kwargs,
         )
         return runner()
+
+    @classmethod
+    @setup_anndata_dsp.dedent
+    def setup_anndata(
+        cls,
+        adata: AnnData,
+        labels_key: str,
+        unlabeled_category: Union[str, int, float],
+        layer: Optional[str] = None,
+        batch_key: Optional[str] = None,
+        size_factor_key: Optional[str] = None,
+        categorical_covariate_keys: Optional[List[str]] = None,
+        continuous_covariate_keys: Optional[List[str]] = None,
+        **kwargs,
+    ):
+        """
+        %(summary)s.
+
+        Parameters
+        ----------
+        %(param_layer)s
+        %(param_batch_key)s
+        %(param_labels_key)s
+        %(param_size_factor_key)s
+        %(param_cat_cov_keys)s
+        %(param_cont_cov_keys)s
+        """
+        setup_method_args = cls._get_setup_method_args(**locals())
+        anndata_fields = [
+            LayerField(REGISTRY_KEYS.X_KEY, layer, is_count_data=True),
+            CategoricalObsField(REGISTRY_KEYS.BATCH_KEY, batch_key),
+            LabelsWithUnlabeledObsField(
+                REGISTRY_KEYS.LABELS_KEY, labels_key, unlabeled_category
+            ),
+            NumericalObsField(
+                REGISTRY_KEYS.SIZE_FACTOR_KEY, size_factor_key, required=False
+            ),
+            CategoricalJointObsField(
+                REGISTRY_KEYS.CAT_COVS_KEY, categorical_covariate_keys
+            ),
+            NumericalJointObsField(
+                REGISTRY_KEYS.CONT_COVS_KEY, continuous_covariate_keys
+            ),
+        ]
+        adata_manager = AnnDataManager(
+            fields=anndata_fields, setup_method_args=setup_method_args
+        )
+        adata_manager.register_fields(adata, **kwargs)
+        cls.register_manager(adata_manager)
